from __future__ import annotations

from textual.widget import Widget

from spiel.transitions.protocol import Direction, Transition


class Swipe(Transition):
    """
    A transition where the current and incoming slide are placed side-by-side
    and gradually slide across the screen,
    with the current slide leaving and the incoming slide entering.
    """

    def initialize(
        self,
        from_widget: Widget,
        to_widget: Widget,
        direction: Direction,
    ) -> None:
        to_widget.styles.offset = ("100%" if direction is Direction.Next else "-100%", 0)

    def progress(
        self,
        from_widget: Widget,
        to_widget: Widget,
        direction: Direction,
        progress: float,
    ) -> None:
        match direction:
            case Direction.Next:
                from_widget.styles.offset = (f"-{progress:.2f}%", 0)
                to_widget.styles.offset = (f"{100 - progress:.2f}%", 0)
            case Direction.Previous:
                from_widget.styles.offset = (f"{progress:.2f}%", 0)
                to_widget.styles.offset = (f"-{100 - progress:.2f}%", 0)



from __future__ import annotations

from typing import Type

from textual.app import ComposeResult
from textual.reactive import reactive

from spiel.screens.screen import SpielScreen
from spiel.slide import Slide
from spiel.transitions.protocol import Direction, Transition
from spiel.triggers import Triggers
from spiel.widgets.fixed_slide import FixedSlideWidget
from spiel.widgets.footer import Footer


class SlideTransitionScreen(SpielScreen):
    DEFAULT_CSS = """\
    SlideTransitionScreen {
        layout: vertical;
        overflow: hidden hidden;
        layers: below above;
    }

    FixedSlideWidget#from {
        layer: above;
    }

    FixedSlideWidget#to {
        layer: below;
    }

    Footer {
        layer: above;
    }

    Footer#dummy {
        layer: below;
    }
    """
    progress = reactive(0, init=False, layout=True)

    def __init__(
        self,
        from_slide: Slide,
        from_triggers: Triggers,
        to_slide: Slide,
        transition: Type[Transition],
        direction: Direction,
    ):
        super().__init__()

        self.from_slide = from_slide
        self.from_triggers = from_triggers
        self.to_slide = to_slide
        self.transition = transition()
        self.direction = direction

    def compose(self) -> ComposeResult:
        from_widget = FixedSlideWidget(self.from_slide, triggers=self.from_triggers, id="from")
        to_widget = FixedSlideWidget(self.to_slide, id="to")

        self.transition.initialize(
            from_widget=from_widget,
            to_widget=to_widget,
            direction=self.direction,
        )

        yield from_widget
        yield to_widget

        yield Footer()
        yield Footer(
            id="dummy"
        )  # a dummy footer to hold space on the "below" layer, won't be displayed

    def watch_progress(self, new_progress: float) -> None:
        from_widget = self.query_one("#from")
        to_widget = self.query_one("#to")

        self.transition.progress(
            from_widget=from_widget,
            to_widget=to_widget,
            direction=self.direction,
            progress=new_progress,
        )



from __future__ import annotations

from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Type, overload

from spiel.slide import Content, Slide
from spiel.transitions.protocol import Transition
from spiel.transitions.swipe import Swipe


@dataclass
class Deck(Sequence[Slide]):
    """
    Represents a "deck" of "slides": a presentation.
    """

    name: str
    """The name of the [`Deck`][spiel.Deck], which will be displayed in the footer."""

    default_transition: Type[Transition] | None = Swipe
    """\
    The default slide transition animation;
    used if the slide being moved to does not specify its own transition.
    Defaults to the [`Swipe`][spiel.Swipe] transition.
    Set to `None` for no transition animation.
    """

    _slides: list[Slide] = field(default_factory=list)

    def slide(
        self,
        title: str = "",
        bindings: Mapping[str, Callable[..., None]] | None = None,
        transition: Type[Transition] | None = None,
    ) -> Callable[[Content], Content]:
        """
        A decorator that creates a new slide in the deck,
        with the decorated function as the [`Slide.content`][spiel.Slide.content].

        Args:
            title: The title to display for the slide.
            bindings: A mapping of
                [keys](https://textual.textualize.io/guide/input/#key)
                to callables to be executed when those keys are pressed,
                when on this slide.
            transition: The transition animation to use when moving to this slide.
                Set to `None` to use the
                [`Deck.default_transition`][spiel.Deck.default_transition]
                of the deck this slide is in.
        """

        def slideify(content: Content) -> Content:
            self.add_slides(
                Slide(
                    title=title,
                    content=content,
                    bindings=bindings or {},
                    transition=transition,
                )
            )
            return content

        return slideify

    def add_slides(self, *slides: Slide) -> None:
        """
        Add `Slide`s to a `Deck`.

        This function is primarily useful when adding multiple slides at once,
        probably generated programmatically.
        If adding a single slide, prefer the [`Deck.slide`][spiel.Deck.slide] decorator.

        Args:
            *slides: The `Slide`s to add.
        """
        self._slides.extend(slides)

    def __len__(self) -> int:
        return len(self._slides)

    @overload
    def __getitem__(self, item: int) -> Slide:
        return self._slides[item]

    @overload
    def __getitem__(self, item: slice) -> Sequence[Slide]:
        return self._slides[item]

    def __getitem__(self, item: int | slice) -> Slide | Sequence[Slide]:
        return self._slides[item]

    def __iter__(self) -> Iterator[Slide]:
        yield from self._slides

    def __contains__(self, item: object) -> bool:
        return item in self._slides



from spiel.app import SuspendType, present
from spiel.constants import __version__
from spiel.deck import Deck
from spiel.slide import Slide
from spiel.transitions.protocol import Direction, Transition
from spiel.transitions.swipe import Swipe
from spiel.triggers import Triggers

__all__ = [
    "Deck",
    "Direction",
    "Slide",
    "SuspendType",
    "Swipe",
    "Transition",
    "Triggers",
    "__version__",
    "present",
]



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from typing import Callable, Mapping, Type

from rich.console import RenderableType
from rich.text import Text

from spiel.transitions.protocol import Transition
from spiel.transitions.swipe import Swipe
from spiel.triggers import Triggers

TRIGGERS = "triggers"

Content = Callable[..., RenderableType]


@dataclass
class Slide:
    """
    Represents a single slide in the presentation.
    """

    title: str = ""
    """The title of the `Slide`, which will be displayed in the footer."""

    content: Content = lambda: Text()
    """\
    A callable that is invoked by Spiel to display the slide's content.

    The function may optionally take arguments with these names:

    - `trigger`: The current [`Trigger`][spiel.Triggers] state, for use in animations.
    """

    bindings: Mapping[str, Callable[..., None]] = field(default_factory=dict)
    """\
    A mapping of
    [keys](https://textual.textualize.io/guide/input/#key)
    to callables to be executed when those keys are pressed,
    when on this slide.
    """

    transition: Type[Transition] | None = Swipe
    """\
    The transition animation to use when moving to this slide.
    Set to `None` to use the
    [`Deck.default_transition`][spiel.Deck.default_transition]
    of the deck this slide is in.
    """

    def render(self, triggers: Triggers) -> RenderableType:
        signature = inspect.signature(self.content)

        kwargs: dict[str, object] = {}
        if TRIGGERS in signature.parameters:
            kwargs[TRIGGERS] = triggers

        return self.content(**kwargs)



from __future__ import annotations

from enum import Enum
from typing import Protocol, runtime_checkable

from textual.widget import Widget


class Direction(Enum):
    """
    An enumeration that describes which direction a slide transition
    animation should move in: whether we're going to the next slide,
    or to the previous slide.
    """

    Next = "next"
    """Indicates that the transition should handle going to the next slide."""

    Previous = "previous"
    """Indicates that the transition should handle going to the previous slide."""


@runtime_checkable
class Transition(Protocol):
    """
    A protocol that describes how to implement a transition animation.

    See [Writing Custom Transitions](./transitions.md#writing-custom-transitions)
    for more details on how to implement the protocol.
    """

    def initialize(
        self,
        from_widget: Widget,
        to_widget: Widget,
        direction: Direction,
    ) -> None:
        """
        A hook function to set up any CSS that should be present at the start of the transition.

        Args:
            from_widget: The widget showing the slide that we are leaving.
            to_widget: The widget showing the slide that we are entering.
            direction: The desired direction of the transition animation.
        """
        ...

    def progress(
        self,
        from_widget: Widget,
        to_widget: Widget,
        direction: Direction,
        progress: float,
    ) -> None:
        """
        A hook function that is called each time the `progress`
        of the transition animation updates.

        Args:
            from_widget: The widget showing the slide that we are leaving.
            to_widget: The widget showing the slide that we are entering.
            direction: The desired direction of the transition animation.
            progress: The progress of the animation, as a percentage
                (e.g., initial state is `0`, final state is `100`).
                Note that this is **not necessarily** bounded between `0` and `100`,
                nor is it necessarily [monotonically increasing](https://en.wikipedia.org/wiki/Monotonic_function),
                depending on the underlying Textual animation easing function,
                which may overshoot or bounce.
                However, it will always start at `0` and end at `100`,
                no matter which `direction` the transition should move in.
        """
        ...
